import moment from "moment";

export type RangeBound = "[]" | "()" | "[)" | "(]"


function parseRange(rangeStr: string) {
	const beg = rangeStr.at(0);
	const end = rangeStr.at(-1);
	if (beg !== "[" && beg !== "(") return null;
	if (end !== "]" && end !== ")") return null;
	const content = rangeStr.slice(1, - 1).trim();
	if (!content.length) return null;
	const items = content.split(",").map(s => s.trim());
	if (items.length !== 2) return null;
	return {
		bound: (beg + end) as RangeBound,
		begin: items[0],
		end: items[1]
	};
}

export abstract class RangeBase<T> {
	#begin: T | null;
	#end: T | null;
	#bound: RangeBound;

	/**
	 * 创建范围，系统会对范围值进行简单校验，如果校验失败则返回null
	 * @param begin 开始值
	 * @param end 结束值
	 * @param bound 边界信息，默认`[)`
	 */
	public static create<T extends RangeBase<any>>(this: new (...args: any) => T, begin: string, end: string, bound?: RangeBound): T | null;
	/**
	 * 创建范围，系统会对范围值进行简单校验，如果校验失败则返回null
	 * @param value 范围值，例如`[1,2)`
	 */
	public static create<T extends RangeBase<any>>(this: new (...args: any) => T, value: string): T | null;
	public static create<T extends RangeBase<any>>(this: new (...args: any) => T, beginOrValue: string, end?: string, bound?: RangeBound): T | null {
		let range: T | null = null;
		if (typeof end === "string") {
			range = new this(beginOrValue, end, bound ?? "[)");
		}
		else {
			const info = parseRange(beginOrValue);
			if (!info) return null;
			range = new this(info.begin, info.end, info.bound);
		}
		if (!range.validate()) return null;
		return range;
	}

	public constructor(begin: T | null, end: T | null, bound: RangeBound) {
		this.#begin = begin;
		this.#end = end;
		this.#bound = bound;
	}

	/** 开始值 */
	public get begin() { return this.#begin; }

	/** 结束值 */
	public get end() { return this.#end; }

	/** 边界信息 */
	public get bound() { return this.#bound; }

	/** 转换为SQL字符串 */
	public toString() {
		const boundBegin = this.bound.at(0);
		const boundEnd = this.bound.at(-1);
		return `'${boundBegin}${this.begin ?? ''},${this.end ?? ''}${boundEnd}'::${this.typename()}`;
	}

	/**
	 * 范围是否包含值
	 * @param value 值
	 */
	public contains(value: T) {
		if (this.begin !== null) {
			if (this.bound.at(0) === "(") {
				if (this.valueCompare(value, this.begin) <= 0) return false;
			}
			else {
				if (this.valueCompare(value, this.begin) < 0) return false;
			}
		}
		if (this.end !== null) {
			if (this.bound.at(-1) === ")") {
				if (this.valueCompare(value, this.end) >= 0) return false;
			}
			else {
				if (this.valueCompare(value, this.end) > 0) return false;
			}
		}
		return true;
	}

	protected abstract validate(): boolean;

	protected abstract typename(): string;

	/**
	 * 值比较， 如果value1大于value2，返回1，如果value1等于value2，返回0，如果value1小于value2，返回-1
	 * @param value1 值1
	 * @param value2 值2
	 */
	protected abstract valueCompare(value1: T, value2: T): number;
}

/**
 * 32位整数范围
 */
export class Int4Range extends RangeBase<number> {
	public constructor(begin: string | null, end: string | null, bound: RangeBound) {
		super(
			begin === null ? null : parseFloat(begin),
			end === null ? null : parseFloat(end),
			bound
		);
	}

	protected validate() { return this.#checkInt(this.begin) && this.#checkInt(this.end); }

	protected typename() { return "int4range"; }

	protected valueCompare(value1: number, value2: number) {
		return value1 - value2;
	}

	#checkInt(value: number | null) {
		if (value === null) return true;
		return Number.isInteger(value);
	}
}

/**
 * 小数范围
 */
export class FloatRange extends RangeBase<number> {
	public constructor(begin: string | null, end: string | null, bound: RangeBound) {
		super(
			begin === null ? null : parseFloat(begin),
			end === null ? null : parseFloat(end),
			bound
		);
	}

	protected validate() { return this.#checkFloat(this.begin) && this.#checkFloat(this.end); }

	protected typename() { return "numrange"; }

	protected valueCompare(value1: number, value2: number) {
		return value1 - value2;
	}

	#checkFloat(value: number | null) {
		if (value === null) return true;
		return Number.isFinite(value);
	}
}

/**
 * 日期范围
 */
export class DateRange extends RangeBase<string> {

	protected validate() { return this.#checkDate(this.begin) && this.#checkDate(this.end); }

	protected typename() { return "daterange"; }

	protected valueCompare(value1: string, value2: string) {
		return moment(value1).diff(value2, "days");
	}

	#checkDate(date: string | null) {
		if (date === null) return true;
		return /^\d{4}-\d{2}-\d{2}$/.test(date);
	}
}

/**
 * 日期时间范围（也就是时间戳范围）
 */
export class DateTimeRange extends RangeBase<string> {
	protected validate() { return this.#checkDatetime(this.begin) && this.#checkDatetime(this.end); }

	protected typename() { return "tsrange"; }

	protected valueCompare(value1: string, value2: string) {
		return moment(value1).diff(value2, "seconds");
	}

	#checkDatetime(datetime: string | null) {
		if (datetime === null) return true;
		return /^\d{4}-\d{2}-\d{2} \d{2}:\d{2}(:\d{2}(\.\d+)?)?$/.test(datetime);
	}
}

